/*
  Tuner support code on OV511

  This code from BT848 driver on FreeBSD

  2002.05.28 H.M
*/

#include <stdio.h>

#include "vid.h"

/* prototype */

int ov511_reg_write(int fd, int reg, int val);
int ov511_i2c_read2(int fd);
int ov511_i2c_write(int fd, int reg, int val);

#define FREQFACTOR  16

#define OFFSET  6.00
#define IF_FREQ 45.75
static int jpnbcst[] = {
	62,     (int)(IF_FREQ * FREQFACTOR),    0,
	13,     (int)(471.25 * FREQFACTOR),     (int)(OFFSET * FREQFACTOR),
	 8,     (int)(193.25 * FREQFACTOR),     (int)(OFFSET * FREQFACTOR),
	 4,     (int)(171.25 * FREQFACTOR),     (int)(OFFSET * FREQFACTOR),
	 1,     (int)( 91.25 * FREQFACTOR),     (int)(OFFSET * FREQFACTOR),
	 0
};
#undef IF_FREQ
#undef OFFSET

#define TBL_CHNL	jpnbcst[ x ]
#define TBL_BASE_FREQ	jpnbcst[ x + 1 ]
#define TBL_OFFSET	jpnbcst[ x + 2 ]
static int
frequency_lookup( int channel )
{
	int	x;

	/* check for "> MAX channel" */
	x = 0;
	if ( channel > TBL_CHNL )
		return( -1 );

	/* search the table for data */
	for ( x = 3; TBL_CHNL; x += 3 ) {
		if ( channel >= TBL_CHNL ) {
			return( TBL_BASE_FREQ +
				 ((channel - TBL_CHNL) * TBL_OFFSET) );
		}
	}

	/* not found, must be below the MIN channel */
	return( -1 );
}
#undef TBL_OFFSET
#undef TBL_BASE_FREQ
#undef TBL_CHNL

int
settv(int fd, int ch)
{
int control = 0xce;
int band;
int frequency, N;

  /* set I2C write slave ID for FI1236MK2(tuner) */
  if(ov511_reg_write(fd, OV511_REG_SID, FI1236MK2_I2C_WRITE_ID) < 0)
    exit(1);
  if(ov511_reg_write(fd, OV511_REG_SRA, FI1236MK2_I2C_READ_ID) < 0)
    exit(1);

  if(ov511_i2c_read2(fd) < 0)
    exit(1);

  frequency = frequency_lookup(ch);

  if ( frequency < (160 * FREQFACTOR  ) )
     band = 0xa0;
  else if ( frequency < (454 * FREQFACTOR ) )
    band = 0x90;
  else
    band = 0x30;

  N = frequency + (int)(45.75 * 16);

fprintf(stderr, "tuner = %02x %02x\n", (N >> 8) & 0x7f, N & 0xff);

  ov511_i2c_write(fd, (N >> 8) & 0x7f, N & 0xff);
//  if(ov511_i2c_read2(fd) < 0)
//    exit(1);

  ov511_i2c_write(fd, control, band);
//  if(ov511_i2c_read2(fd) < 0)
//    exit(1);

  /* reset the OV511 */
  if(ov511_reg_write(fd, OV511_REG_RST, 0x03) < 0)
    exit(1);

return 1;
}
